
# THIS FILE IS GENERATED FROM ORANGE SETUP.PY
short_version = '3.36.2'
version = '3.36.2'
full_version = '3.36.2'
git_revision = '61033eccedf8143804d7010ffe3621f401e6cb3c'
release = True

if not release:
    version = full_version
    short_version += ".dev"
